﻿#include "define.h"

// IN
uniform int   layer;
uniform bool  closest;
uniform bool  correctAlpha;
uniform float transparency;
uniform bool  useTransparency;
uniform bool  useTranslucency;
uniform vec4  color_background;

#if multisample
	layout(binding = 0, r32ui) readonly uniform uimage2DMS	    image_counter;
#if read_only
	layout(binding = 1, rg32f) readonly uniform  image2DMSArray image_peel;
#else
	layout(binding = 1, rg32f) coherent uniform  image2DMSArray image_peel;
#endif 
#else
	layout(binding = 0, r32ui) readonly uniform uimage2DRect	image_counter;
#if read_only
	layout(binding = 1, rg32f) readonly uniform  image2DArray	image_peel;
#else
	layout(binding = 1, rg32f) coherent uniform  image2DArray	image_peel;
#endif 
#endif 

// OUT
layout(location = 0, index = 0) out  vec4 out_frag_color;

	void bubbleSort		  (const ivec2 coords, const int num);
	void swap			  (const ivec2 coords, const int a, const int b);
	void siftDown		  (const ivec2 coords, int parent, int end);
	void heapSort		  (const ivec2 coords, const int num);

	vec4 resolveClosest	  (const ivec2 coords, const int num);
	vec4 resolveAlphaBlend(const ivec2 coords, const int num);

void main(void)
{
	int   counter;
	ivec2 coords = ivec2(gl_FragCoord.xy);

#if multisample
	counter = int(imageLoad(image_counter, coords, gl_SampleID).r);
#else
	counter = int(imageLoad(image_counter, coords).r);
#endif 

	if(counter > 0)
	{

		vec4 c;
		if(useTransparency || useTranslucency)
		{
			bubbleSort(coords, counter);
			c = resolveAlphaBlend(coords, counter);
		}
		else
		{
			if(closest)
			{
				out_frag_color = resolveClosest (coords, counter); return;
			}
			if(layer+1 > counter)
				discard;
			
			bubbleSort(coords, counter);
#if multisample
			c  = unpackUnorm4x8(uint(imageLoad(image_peel, ivec3(coords, layer), gl_SampleID).r)).abgr;
#else
			c  = unpackUnorm4x8(uint(imageLoad(image_peel, ivec3(coords, layer)).r)).abgr;
#endif
		}
		out_frag_color   = c;
	}
	else
		discard;
}
vec4 resolveAlphaBlend(const ivec2 coords, const int num)
{
	vec2 Ci;
	
	vec4 C0=vec4(0.0f), color, finalColor=vec4(0.0f);
	float thickness=0.0f, prevZ = 0.0f, thicknessAlpha=0.0f, Zi1;
	
	for(int i=0; i<num; i++)
	{
#if multisample
		Ci = imageLoad(image_peel, ivec3(coords, i), gl_SampleID).rg;
#else
		Ci = imageLoad(image_peel, ivec3(coords, i)).rg;
#endif
		if(i==0)
		{
			C0 = unpackUnorm4x8(uint(Ci.r)).abgr;
			if(correctAlpha) thicknessAlpha = Ci.g*0.5f;
		}

		if(useTranslucency)
		{
			if(i%2==1) thickness += Ci.g - prevZ;
			prevZ = Ci.g;
		}
		else
		{
			color = vec4(unpackUnorm4x8(uint(Ci.r)).abg, transparency);
			if(correctAlpha)
			{
				if(i%2 == num%2)
				{
#if multisample
					Zi1 = imageLoad(image_peel, ivec3(coords, i+1), gl_SampleID).g;
#else
					Zi1 = imageLoad(image_peel, ivec3(coords, i+1)).g;
#endif
					Zi1 = abs(Zi1);
					thicknessAlpha = (Zi1-Ci.g)*0.5f;
				}
				color.w		= 1.0f - pow(1.0f - color.w, thicknessAlpha*sigma);
				color.rgb  *= color.a;
			}
			finalColor += color*(1.0f-finalColor.a);
		}
	}

	if(useTransparency) finalColor += color_background*(1.0f-finalColor.a);
	else  
	{
		const float Ka    = 0.8f;
		const vec4  jade  = vec4(0.4, 0.14, 0.11, 1.0)*8.0f;
		const vec4  green = vec4(0.3, 0.7 , 0.1 , 1.0)*4.0f;
		
		finalColor = (Ka*exp(-sigma*thickness)*vec4(1.0f) + C0)*jade;	
	}

	return finalColor;
}

vec4 resolveClosest(const ivec2 coords, const int num)
{
	vec2 Ci;
	vec2 maxC = vec2(0.0f,1.0f);

	for(int i=0; i<num; i++)
	{
#if multisample
		Ci = imageLoad(image_peel, ivec3(coords, i), gl_SampleID).rg;
#else
		Ci = imageLoad(image_peel, ivec3(coords, i)).rg;
#endif
		if(Ci.g < maxC.g)
			maxC = Ci;
	}
	return unpackUnorm4x8(uint(maxC.r)).abgr;
}

void bubbleSort(const ivec2 coords, const int num)
{
	vec4 c0,c1;

	for (int i = (num - 2); i >= 0; --i)
	{
		for (int j = 0; j <= i; ++j)
		{
#if multisample
			c0 = imageLoad(image_peel, ivec3(coords, j  ), gl_SampleID);
			c1 = imageLoad(image_peel, ivec3(coords, j+1), gl_SampleID);
#else
			c0 = imageLoad(image_peel, ivec3(coords, j));
			c1 = imageLoad(image_peel, ivec3(coords, j+1));
#endif
			if (c0.g > c1.g)
			{
#if multisample
				imageStore(image_peel, ivec3(coords, j  ), gl_SampleID, c1);
				imageStore(image_peel, ivec3(coords, j+1), gl_SampleID, c0);
#else
				imageStore(image_peel, ivec3(coords, j  ), c1);
				imageStore(image_peel, ivec3(coords, j+1), c0);
#endif
			}
		}
	}
}