﻿#include "define.h"

// IN
uniform int	   size;
uniform int	   layer;
uniform bool   closest;
uniform bool   correctAlpha;
uniform bool   useTransparency;
uniform bool   useTranslucency;
uniform vec4   color_background;
uniform float  transparency;

#if multisample
	layout(binding = 0, r32ui) readonly uniform uimage2DMS    image_counter;
	layout(binding = 1, r32ui) readonly uniform uimage2DMS    image_next;
#else
	layout(binding = 0, r32ui) readonly uniform uimage2DRect  image_counter;
	layout(binding = 1, r32ui) readonly uniform uimage2DRect  image_next;
#endif 
	layout(binding = 2, rg32f) coherent uniform imageBuffer   image_peel;

	vec4 sharedPoolGetValue	  (const uint index				   ) {return imageLoad (image_peel, (int)index	   );}
	void sharedPoolSetValue	  (const uint index, const vec4 val) {		 imageStore(image_peel, (int)index, val);}

#if multisample
	uint getPixelNextAddress  (const ivec2 coords)	{return imageLoad(image_next   , coords, gl_SampleID).x-1U;}
	uint getPixelFragCounter  (const ivec2 coords)	{return imageLoad(image_counter, coords, gl_SampleID).x;   }
#else
	uint getPixelNextAddress  (const ivec2 coords)	{return imageLoad(image_next   , coords).x-1U;}
	uint getPixelFragCounter  (const ivec2 coords)	{return imageLoad(image_counter, coords).x;   }
#endif
	
	vec4 resolveAlphaBlend(const uint page_id, const uint num);
	vec4 resolveClosest	  (const uint page_id, const uint num);
	void bubbleSort		  (const uint page_id, const uint num);

// OUT
layout(location = 0, index = 0) out  vec4 out_frag_color;

void main(void)
{
	ivec2 coords = ivec2(gl_FragCoord.xy);

	uint counter = getPixelFragCounter(coords);
	if(counter > 0U)
	{
		if(counter > size)
			counter = size;
		
		uint page_id = getPixelNextAddress(coords);

		vec4 c = vec4(0.0f);

		if(useTransparency || useTranslucency)
		{
			bubbleSort(page_id, counter);
			c = resolveAlphaBlend(page_id, counter);
		}
		else
		{
			if(closest)
			{
				out_frag_color = resolveClosest(page_id, counter);
				return;
			}

			if(counter <= layer)
				discard;

			bubbleSort(page_id, counter);

			for (uint i = 0; i < counter; i++)
			{
				if (i == layer)
				{
					c = sharedPoolGetValue(page_id);
					c = unpackUnorm4x8(uint(c.r)).abgr;
					break;
				}
				page_id--;
			}
		}

		out_frag_color = c;
	}
	else
		discard;
}


vec4 resolveAlphaBlend(const uint page_id, const uint num)
{
	float thickness=0.0f, thicknessAlpha=0.0f, prevZ = 0.0f, Zi1;
	vec4  finalColor=vec4(0.0f), color,C0=vec4(0.0f),Ci;
	uint  cur_page_id = page_id;

	for (uint i=0U; i < num; i++)
	{
		Ci = sharedPoolGetValue(cur_page_id);
		if(i == 0U)
		{
			C0 = unpackUnorm4x8(uint(Ci.r)).abgr;
			if(correctAlpha) thicknessAlpha = Ci.g*0.5f;
		}	

		if(useTranslucency)
		{
			if(i%2 == 1) thickness += Ci.g - prevZ;
			prevZ = Ci.g;
		}
		if(useTransparency)
		{
			color= vec4(unpackUnorm4x8(uint(Ci.r)).abg, transparency);
			if(correctAlpha)
			{
				if(i%2 == num%2)
				{
					Zi1 = sharedPoolGetValue(cur_page_id-1U).g;
					thicknessAlpha=(Zi1-Ci.g)*0.5f;
				}
				color.w=1.0f-pow(1.0f-color.w, thicknessAlpha*sigma);
				color.rgb *= color.a;
			}
			finalColor = finalColor + color*(1.0f-finalColor.a);
		}
		cur_page_id--;
	}
	
	if(useTransparency) finalColor += color_background*(1.0f-finalColor.a);
	else
	{
		const float Ka    = 0.8f;
		const vec4  jade  = vec4(0.4, 0.14, 0.11, 1.0)*8.0f;
		const vec4  green = vec4(0.3, 0.7 , 0.1 , 1.0)*4.0f;
		
		finalColor = (Ka*exp(-sigma*thickness)*vec4(1.0f) + C0)*jade;	
	}
	return finalColor;
}

vec4 resolveClosest(const uint page_id, const uint num)
{
	uint cur_page_id = page_id;
	vec2 value, minFrag = vec2(0.0f, 1.0f);

	for (uint i = 0; i < num; i++)
	{
		value = sharedPoolGetValue(cur_page_id).rg;
		if(value.g < minFrag.g)
			minFrag = value;
		cur_page_id--;
	}
	return unpackUnorm4x8(uint(minFrag.r)).abgr;
}

void bubbleSort(const uint page_id, const uint num)
{
	vec4 c0, c1;
	uint p0, p1;
	int  k = int(num) - 2;

	for (int i = k; i >= 0; --i)
	{
		p0 = page_id;
		for (int j = 0; j <= i; ++j)
		{		
			p1 = p0-1U;
			c0 = sharedPoolGetValue(p0);
			c1 = sharedPoolGetValue(p1);
			if (c0.g > c1.g)
			{
				sharedPoolSetValue (p0, c1);
				sharedPoolSetValue (p1, c0);
			}
			p0 = p1;
		}
	}
}